
#    MIDAS Access Server  for  MEIS EPICS system

global env
source [file join $env(MIDASBASE) TclHttpd tcl MIDASDomainHandler.tcl]       ;#  load default support procedures

source [file join $env(MIDASBASE) MIDASpackages EPICSAccess1.0 pkgIndex.tcl]
package require EPICSAccess

global MIDAS_EPICSAccessServer_Verbose
set MIDAS_EPICSAccessServer_Verbose 0

global MIDAS_EPICSChIDStore
set MIDAS_EPICSChIDStore(dummy) ""

# Use namespaces to isolate our methods
namespace eval urn:EPICSAccessServer {

    proc AccessServerVerbose {opt} {
        do_AccessServerVerbose $opt
        return [rpcvar::rpcvar int 0]
    }

    proc CreateChannel {PVname} {
        set chid [do_CreateChannel $PVname]
        return [rpcvar::rpcvar int $chid]
    }

    proc ClearChannel {PVname} {
        set rc [do_ClearChannel $PVname]
        return [rpcvar::rpcvar int $rc]
    }

    proc GetInt {PVname} {
        set rc [do_GetInt $PVname]
        return [rpcvar::rpcvar int $rc]
    }

    proc PutInt {PVname vl} {
        set rc [do_PutInt $PVname $vl]
        return [rpcvar::rpcvar int $rc]
    }

    proc GetFloat {PVname} {
        set rc [do_GetFloat $PVname]
        return [rpcvar::rpcvar float $rc]
    }

    proc PutFloat {PVname vl} {
        set rc [do_PutFloat $PVname $vl]
        return [rpcvar::rpcvar int $rc]
    }

    proc GetString {PVname} {
        set rc [do_GetString $PVname]
        return [rpcvar::rpcvar string $rc]
    }

    proc PutString {PVname vl} {
        set rc [do_PutString $PVname $vl]
        return [rpcvar::rpcvar int $rc]
    }

    proc do_AccessServerVerbose {option} {
        global MIDAS_EPICSAccessServer_Verbose
        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method AccessServerVerbose with parameter $option"}
        set MIDAS_EPICSAccessServer_Verbose $option
        EPICSAccess_Verbose $option
    }

    proc do_CreateChannel {PVname} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method CreateChannel with parameter $PVname"}

        if {[info exists ChIDStore($PVname)]} {return $ChIDStore($PVname)}

        set z [catch {EPICS_CreateChannel $PVname} m]
        if {$z == 0} then {
           set ChIDStore($PVname) $m
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_ClearChannel {PVname} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method ClearChannel with parameter $PVname"}

        if {![info exists ChIDStore($PVname)]} {return 0}

        set z [catch {EPICS_ClearChannel $ChIDStore($PVname)} m]
        if {$z == 0} then {
           unset ChIDStore($PVname)
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_PutInt {PVname vl} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method PutInt with parameters $PVname $vl"}

        if {![info exists ChIDStore($PVname)]} {return -code error -errorcode 0x10004 -errorinfo "unknown variable"}

        set z [catch {EPICS_PutInt $ChIDStore($PVname) $vl} m]
        if {$z == 0} then {
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_GetInt {PVname} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method GetInt with parameter $PVname"}

        if {![info exists ChIDStore($PVname)]} {return -code error -errorcode 0x10004 -errorinfo "unknown variable"}

        set z [catch {EPICS_GetInt $ChIDStore($PVname)} m]
        if {$z == 0} then {
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_PutFloat {PVname vl} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method PutFloat with parameters $PVname $vl"}

        if {![info exists ChIDStore($PVname)]} {return -code error -errorcode 0x10004 -errorinfo "unknown variable"}

        set z [catch {EPICS_PutFloat $ChIDStore($PVname) $vl} m]
        if {$z == 0} then {
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_GetFloat {PVname} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method GetFloat with parameter $PVname"}

        if {![info exists ChIDStore($PVname)]} {return -code error -errorcode 0x10004 -errorinfo "unknown variable"}

        set z [catch {EPICS_GetFloat $ChIDStore($PVname)} m]
        if {$z == 0} then {
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_PutString {PVname vl} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method PutString with parameters $PVname $vl"}

        if {![info exists ChIDStore($PVname)]} {return -code error -errorcode 0x10004 -errorinfo "unknown variable"}

        set z [catch {EPICS_PutString $ChIDStore($PVname) $vl} m]
        if {$z == 0} then {
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }

    proc do_GetString {PVname} {
        global errorCode MIDAS_EPICSAccessServer_Verbose
        upvar #0 MIDAS_EPICSChIDStore ChIDStore

        if {$MIDAS_EPICSAccessServer_Verbose} {puts "called method GetString with parameter $PVname"}

        if {![info exists ChIDStore($PVname)]} {return -code error -errorcode 0x10004 -errorinfo "unknown variable"}

        set z [catch {EPICS_GetString $ChIDStore($PVname)} m]
        if {$z == 0} then {
           return $m
        } else {
           return -code error -errorcode 0x10005 -errorinfo "received error response: rc=$z\n $m\n $errorCode"
        }
    }




    # We have to publish the public methods...

     SOAP::export AccessServerVerbose
     SOAP::export CreateChannel
     SOAP::export ClearChannel
     SOAP::export PutInt
     SOAP::export GetInt
     SOAP::export PutFloat
     SOAP::export GetFloat
     SOAP::export PutString
     SOAP::export GetString


}

# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /EPICSAccessServer -namespace urn:EPICSAccessServer -uri urn:EPICSAccessServer]


proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/EPICSAccessServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}
